'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _webpackSources = require('webpack-sources');

var _varname = require('varname');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

module.exports = WebpackSystemRegister;

function WebpackSystemRegister(options) {
	if ((typeof options === 'undefined' ? 'undefined' : _typeof(options)) !== 'object') throw new Error("webpack-system-register takes exactly one argument (pass an options object)");

	if (options.systemjsDeps && !Array.isArray(options.systemjsDeps)) throw new Error('webpack-system-register requires that systemjsDeps is an array of strings');

	this.options = _extends({
		minify: false,
		systemjsDeps: []
	}, options);
}

WebpackSystemRegister.prototype.apply = function (compiler) {
	var options = this.options;
	options.systemjsDeps = (options.systemjsDeps || []).map(function (depName) {
		if (typeof depName === 'string') {
			// literal string match, as a regular expression
			return new RegExp('^' + depName + '$');
		} else if (depName instanceof RegExp) {
			return depName;
		} else {
			throw new Error('systemJsDeps must either be a string or a regular expression');
		}
	});

	var externalModuleFiles = [];
	var externalModules = [];

	if (!compiler.options.resolve) {
		compiler.options.resolve = {};
	}

	if (!compiler.options.resolve.alias) {
		compiler.options.resolve.alias = {};
	}

	compiler.plugin('normal-module-factory', function (nmf) {
		nmf.plugin('before-resolve', function (result, callback) {
			if (!result) {
				return callback();
			}

			if (options.systemjsDeps.find(function (dep) {
				return dep.test(result.request);
			})) {
				(function () {
					var filename = 'node_modules/__' + toJsVarName(result.request);
					if (externalModuleFiles.indexOf(filename) < 0) {
						externalModuleFiles.push(filename);
						_fs2.default.writeFile(filename, 'module.exports = ' + toJsVarName(result.request) + ';', function (err) {
							if (err) {
								console.error(err);
								throw err;
							}

							externalModules.push({
								depFullPath: result.request,
								depVarName: toJsVarName(result.request)
							});
							result.request = _path2.default.resolve(process.cwd(), filename);

							callback(null, result);
						});
					} else {
						result.request = _path2.default.resolve(process.cwd(), filename);
						callback(null, result);
					}
				})();
			} else {
				callback(null, result);
			}
		});

		nmf.plugin('after-resolve', function (result, callback) {
			if (!result) {
				return callback();
			}

			if (options.systemjsDeps.find(function (dep) {
				return dep.test(result.request);
			})) {
				result.resource = _path2.default.resolve(process.cwd(), '__' + toJsVarName(result.request));
			}

			callback(null, result);
		});
	});

	compiler.plugin("compilation", function (compilation) {

		// http://stackoverflow.com/questions/35092183/webpack-plugin-how-can-i-modify-and-re-parse-a-module-after-compilation
		compilation.plugin('seal', function () {
			compilation.modules.forEach(function (module) {
				var isEntry = module.entry;
				var entries = compiler.options.entry || {};
				for (var entryName in entries) {
					isEntry = module.rawRequest === entries[entryName];
				}
				if (isEntry && module._source) {
					module._source._value += '\n$__register__main__exports(exports);';
				}
			});
		});

		// Based on https://github.com/webpack/webpack/blob/ded70aef28af38d1deb2ac8ce1d4c7550779963f/lib/WebpackSystemRegister.js
		compilation.plugin("optimize-chunk-assets", function (chunks, callback) {
			chunks.forEach(function (chunk) {
				if (!chunk.initial) {
					return;
				}

				chunk.files.forEach(function (file) {
					compilation.assets[file] = new _webpackSources.ConcatSource(sysRegisterStart(options, externalModules), compilation.assets[file], sysRegisterEnd(options));
				});
			});
			callback();
		});
	});
};

function sysRegisterStart(opts, externalModules) {
	var result = 'System.register(' + registerName() + depsList() + ', function($__export) {\n  var ' + externalModules.map(toDepVarName).map(toCommaSeparatedList).reduce(toString) + ';\n\n  function $__register__main__exports(exports) {\n    for (var exportName in exports) {\n\t  $__export(exportName, exports[exportName]);\n    }\n  }\n\n  function $__wsr__interop(m) {\n\treturn m.__useDefault ? m.default : m;\n  }\n\n  return {\n    setters: [' + externalModules.map(toDepVarName).map(toSetters.bind(null, opts)).reduce(toString) + '\n    ],\n    execute: function() {\n';

	return opts.minify ? minify(result) : result;

	function registerName() {
		return opts.registerName ? '\'' + opts.registerName + '\', ' : '';
	}

	function depsList() {
		return '[' + externalModules.map(toDepFullPath).map(toStringLiteral).map(toCommaSeparatedList).reduce(toString) + ']';
	}

	function toCommaSeparatedList(name, i) {
		return '' + (i > 0 ? ', ' : '') + name;
	}

	function toSetters(opts, name, i) {
		// webpack needs the __esModule flag to know how to do it's interop require default func
		var result = (i > 0 ? ',' : '') + '\n      function(m) {\n        ' + name + ' = $__wsr__interop(m);\n      }';

		return opts.minify ? minify(result) : result;
	}

	function toStringLiteral(str) {
		return '\'' + str + '\'';
	}

	function toString(prev, next) {
		return prev + next;
	}
}

function sysRegisterEnd(opts) {
	var result = '\n    }\n  }\n});\n';
	return opts.minify ? minify(result) : result;
}

function toJsVarName(systemJsImportName) {
	return (0, _varname.camelcase)(removeSlashes(moduleName(systemJsImportName)));
}

function moduleName(systemJsImportName) {
	return systemJsImportName.includes('!') ? systemJsImportName.slice(0, systemJsImportName.indexOf('!')) : systemJsImportName;
}

function removeSlashes(systemJsImportName) {
	return systemJsImportName.replace('/', '');
}

function toDepVarName(externalModule) {
	return externalModule.depVarName;
}

function toDepFullPath(externalModule) {
	return externalModule.depFullPath;
}

function minify(string) {
	return string.replace(/\n/g, '');
}